
// Application CompanyInfo - Contains info about a company
//
//  Public Interfaces:
//
//    CompanyInfo() - Constructor


//    set(value, attribute) - set a value for an attribute
//    set(value, attribute, year) - set a value for an attribute for a
//		specified year
//    set(value, attribute, year, per) - set a value for an attribute for
//		a specified year and period
//    set(array_of_items) - set a collection of values
//		Each element of the array is itself an array of
//		    2 items - value & attribute
//		    3 items - value, attribute & year
//		or  4 items - value, attribute, year & period

//    get(attribute) - get a value for an attribute
//    get(attribute, year) - get a value for an attribute for a
//		specified year
//    get(attribute, year, per) - get a value for an attribute for
//		a specified year and period

//    dump() - for debugging; dump all settings out

// ----------------------------------------------------------------

//  All values & attributes are arbitrary content strings; e.g. "34.5",
//    "Outstanding Shares", etc.
//  Years are strings of the year number, e.g. "1998" etc.
//  Periods are strings made up a letter and a number (case is not relevant);
//    e.g. "Q1", "m3" - 1st quarter, third month, etc.
//    With the allowable indicator letters as:
//	Q - Quarter (1-4)
//	H - Half (1-2)
//	M - Month (1-12)
//	W - Week (1-52)
//	D - Day (1-365)

package sptossg ;

import java.util.*;
import java.io.*;

public class CompanyInfo implements CompanyStorage {



    // ****************************************************************
    // Data structures initialization & processing

    // Constructor
    public CompanyInfo () {
    }

    // Storage for all static values, i.e. values not characterized by a
    // date; e.g. company name, ticker, etc.
    Hashtable staticvals = new Hashtable();

    // Storage for all datasets, collections of data characterized by
    // dates; e.g. annual sales, quarterly dividends, etc.
    //
    // the key to this hashtable is a string representing the attribute
    // with the value being the DataSet for that attribute
    Hashtable datasets = new Hashtable();

    // getDataSet - retrieve the dataset for a specified attribute.  If a
    // dataset hasn't yet been created for the attribute, it will create it
    // if build is true.  If there is a defined DynamicDataSet (a DataSet
    // with a formula to calculate the value) then it will add a copy to
    // the company.
    DataSet getDataSet(String str, boolean build) {
	Object set;

	// First check for any instantiated dataset
	set = datasets.get(str);

        // next, get any uninstantiated DynamicDataSet and instantiate with
        // the current company
	if (set == null) {
	    set = DynamicDataSet.find(str);
	    if (set != null) {
		DynamicDataSet ds = (DynamicDataSet) set;
		ds = ds.instantiate(this);
		if (baseperiod != null) { ds.setBasePeriod(baseperiod); }
		datasets.put(str, ds);
		set = ds;
	    }
	}

	// otherwise build a regular DataSet
	if (set == null && build == true) {
	    DataSet ds = new DataSet(str);
	    datasets.put(str, ds);
	    set = ds;
	}

	if (set instanceof DataSet) {
	    return (DataSet) set;
	} else {
	    // throw new incompatibleFormatException();
	    return  null;
	}
    }

    Period baseperiod = null;

    // set the base period for attributes of a specific frequency to handle
    // relative periods properly
    public void setBasePeriod(int year, int yearpart, int frequency) {
	baseperiod = new Period(year, yearpart, frequency);
	Enumeration e;

	e = datasets.keys();
	while (e.hasMoreElements()) {
	    String key = (String) e.nextElement();
	    DataSet ds = (DataSet) datasets.get(key);
	    if (ds.frequency() == frequency) {
		ds.setBasePeriod(baseperiod);
	    }
	}
    }

    // ****************************************************************
    // Storage methods

    // value set for static values
    public String set(String value, String attr) {
	staticvals.put(attr, value);
	return value;
    }

    // value set for annual data
    public String set(String value, String attr, String year) {
	return set(value, attr, year, "Y1");
    }

    // value set for periodic data (quarterly, etc.), using string parameters
    public String set(String value, String attr, String year, String yearpart) {
	DataSet ds = getDataSet(attr, true);
	ds.store(value, year, yearpart);
	return value;
    }

    // value set for periodic data, using numeric parameters
    public String set(String value, String attr, int year, int yearpart, int frequency) {
	DataSet ds = getDataSet(attr, true);
	ds.store(value, year, yearpart, frequency);
	return value;
    }

    // value set for array of data (attr, year, period), all items as strings
    public String set(String[][] items) {

	for (int i = 0; i < items.length; i++) {
	    String itm[] = items[i];

	    String val = itm[0];
	    String label = itm[1];

	    if (itm.length == 2) {
		set(val, label);
	    } else if (itm.length == 3) {
		String yr = itm[2];
		set(val, label, yr);
	    } else {
		String yr = itm[2];
		String per = itm[3];
		set(val, label, yr, per);
	    }
	}
	return "";
    }

    // ****************************************************************
    // Retrieval methods

    // value retrieval for static data
    public String get(String attr) {
	return (String) staticvals.get(attr);
    }

    // value retrieval for annual data
    public String get(String attr, String year) {
	return get(attr, year, "Y1");
    }

    // value retrieval for periodic data, string parameters
    public String get(String attr, String year, String yearpart) {
	String retval = null;
	DataSet ds = getDataSet(attr, false);
	if (ds != null) {
	    retval = ds.retrieve(year, yearpart);
	}
	return retval;
    }

    // value retrieval for periodic data, numeric parameters
    public String get(String attr, int year, int yearpart, int frequency) {
	String retval = null;
	DataSet ds = getDataSet(attr, false);
	if (ds != null) {
	    retval = ds.retrieve(year, yearpart, frequency);
	}
	return retval;
    }

    // value retrieval for periodic data, Period parameter
    public String get(String attr, Period per) {
	String retval = null;
	DataSet ds = getDataSet(attr, false);
	if (ds != null) {
	    retval = ds.retrieve(per);
	}
	return retval;
    }

    // ****************************************************************
    // Routines for testing

    void dumpAttrs() {
	Enumeration e;

	int i = 0;
	e = staticvals.keys();
	System.out.println("Static Values: ");
	while (e.hasMoreElements()) {
	    String key = (String) e.nextElement();
	    System.out.println(i++ + ": " + key);
	}

	e = datasets.keys();
	System.out.println("Periodic Values: ");
	while (e.hasMoreElements()) {
	    String key = (String) e.nextElement();
	    DataSet ds = (DataSet) datasets.get(key);
	    int freq = ds.frequency();
	    String freqstr;
	    if (freq == 0) {
		freqstr = "unknown";
	    } else {

		freqstr = Period.periodicString(freq);
	    }
	    System.out.println(i++ + ": (" + freqstr + ") " + key);
	}
    }

    // test_sets - a method demonstrating how to use the various public
    // interface methods (set, get) to store and retrieve values
    void test_sets() {
	// Example SETs, done individually
	set("SSG Creators, Inc.", "Name");
	set("58.6", "Annual Sales", "2001");
	set("10.2", "Quarterly Income", "1999", "Q3");

	// Example SETs, via array
	String tostore[][] = {
	    {"BIBH", "Ticker"},
	    {"47.1", "Annual Sales", "1999"},
	    {"53.4", "Annual Sales", "2000"},
	    {"92.0", "Annual Sales", "2002"},

	    {"11.3", "Annual Profit", "1999"},
	    {"19.4", "Annual Profit", "2000"},
	    {"24.6", "Annual Profit", "2001"},
	    {"44.4", "Annual Profit", "2002"},

	    {"8.9", "Quarterly Income", "1999", "Q1"},
	    {"9.4", "Quarterly Income", "1999", "Q2"},
	    {"11.4", "Quarterly Income", "1999", "Q4"},

	    {"18.9", "Quarterly Income", "2000", "Q1"},
	    {"19.4", "Quarterly Income", "2000", "Q2"},
	    {"20.2", "Quarterly Income", "2000", "Q3"},
	    {"21.4", "Quarterly Income", "2000", "Q4"},
	};
	set(tostore);


	// Example GETs
	String name = get("Name");
	System.out.println("Name = '" + name + "'.");

	String sales2000 = get("Annual Sales", "2000");
	System.out.println("Sales(2000) = '" + sales2000 + "'.");

	String qi2000q1 = get("Quarterly Income", "2000", "Q1");
	System.out.println("Quarterly Income(2000,Q1) = '" + qi2000q1 + "'.");

	// Some more GETs, by array; some aren't set
	String tofetch[][] = {
	    {"Annual Sales", "2001"},
	    {"Quarterly Income", "1999", "Q2"},
	    {"Quarterly Income", "1989", "Q1"},
	    {"Annual Profit", "2002"},
	    {"Annual Profit", "2004"}};

	for (int i = 0; i < tofetch.length; i++) {
	    String itm[] = tofetch[i];

	    String label = itm[0];
	    String yr = itm[1];

	    if (itm.length == 2) {
		String val = get(label, yr);
		System.out.println(label + " for " + yr + " = " + val);
	    } else {
		String per = itm[2];

		String val = get(label, yr, per);
		System.out.println(label + " for " + yr + "/" + per + " = " + val);
	    }
	}

	System.out.println();

	dump();
    }

    public Vector getKeys(String attr) {
	DataSet ds = (DataSet) datasets.get(attr);
	if (ds != null) { return ds.getKeys(this); }
	return null;
    }

    public void dump(String attr) {
	DataSet ds = (DataSet) datasets.get(attr);
	if (ds == null) {
	    Debug.debug("No Regular dataset for " + attr);
	    ds = DynamicDataSet.find(attr);
	    if (ds != null) {
	    Debug.debug("Found Dynamic Dataset for " + attr);
		DynamicDataSet dds = (DynamicDataSet) ds;
		ds = dds.instantiate(this);
		if (baseperiod != null) { ds.setBasePeriod(baseperiod); }
	    }
	}
	if (ds != null) {
	    ds.dump(this, attr);
	} else {
	    Debug.debug("No dataset found for " + attr);
	}
    }

    // dump - print out all attributes and their values
    public void dump() {

	System.out.println("Static Values: ");
	Enumeration e = staticvals.keys();
	while (e.hasMoreElements()) {
	    String key = (String) e.nextElement();
	    String value = (String) staticvals.get(key);
	    if (value != null && ! value.equals("")) {
		System.out.println("    " + key + " = " + value);
	    }
	}

	System.out.println("Periodic Values: ");
	e = datasets.keys();
	while (e.hasMoreElements()) {
	    dump((String) e.nextElement());
	}
    }

    // ****************************************************************
    // Data file format descriptions


    // fileFormatItem - an inner class that defines an item that goes into
    // a data output file.  There are three types of items;
    //   CONSTANTS - A string written into the file as is;
    //      e.g. "Version 7", "," etc.
    //   a STATIC - a non-dated data item, as stored in the table STATICVALS
    //      e.g. "Company Name", "Company Ticker"
    //   a PERIODIC - a dated item, relative to the most recent date stored;
    //      e.g. "Annual Sales" for 3 years ago, "Quarerly EPS" for 8 quarters ago;
    //
    // each item contains a count, interpreted as follows:
    //  for CONSTANTS or STATICS - value is inserted N times
    //  for PERIODICS - the OFFSET specifies the start period; the INCREMENT specifies
    //   how to get the next one;

    // CALCULATED may be used later for attributes that are calculated
    // from other attributes
    // static public final int CALCULATED = 0;

    // CONSTANT values are specified explicitly, not retrieved based on company
    static public final int CONSTANT = 0;


    // STATIC (or SINGLE) values have one value per company
    static public final int STATIC = 1;
    static public final int SINGLE = 1;

    // PERIODIC data have values associated with date periods (years,
    // quarters, etc.)
    static public final int PERIODIC = 2;
    static public final int ANNUAL = 3;
    static public final int QUARTERLY = 4;

    static class fileFormatItem {

	String label;
	int itmType;
	int count;

	int frequency;

	int offset;
	int increment;

	public fileFormatItem(String label) {
	    this(label, SINGLE);
	};

	public fileFormatItem(String label, int itmType) {
	    this(label, itmType, 1, 0, 0);
	};

        public fileFormatItem(String label, int itmType, int count, int
			      offset, int increment) {
	    this.label = label;
	    if (itmType == ANNUAL) {
		this.frequency = 1;
		this.itmType = PERIODIC;
	    } else if (itmType == QUARTERLY) {
		this.frequency = 4;
		this.itmType = PERIODIC;
	    } else {
		this.itmType = itmType;
	    }
	    this.count = count;
	    this.offset = offset;
	    this.increment = increment;
	}
    }

    // This array describes the format of an SSG type 1 file, made up of
    // 400 entries.
    static fileFormatItem format_SSG1[] = {

	new fileFormatItem("High Price", ANNUAL, 10, -9, 1),
	new fileFormatItem("Low Price", ANNUAL, 10, -9, 1),
	new fileFormatItem("Sales", ANNUAL, 10, -9, 1),
	new fileFormatItem("Tax Rate", ANNUAL, 10, -9, 1),
	new fileFormatItem("PreTax Inc", ANNUAL, 10, -9, 1),
	new fileFormatItem("Net Income", ANNUAL, 10, -9, 1),
	new fileFormatItem("Book Value", ANNUAL, 10, -9, 1),
	new fileFormatItem("Earnings per Share", ANNUAL, 10, -9, 1),
	new fileFormatItem("Dividend per Share", ANNUAL, 10, -9, 1),

	// 90 items, annual data

	new fileFormatItem("Current Price"),
	new fileFormatItem("Current P/E"),
	new fileFormatItem("Current Dividend"),
	new fileFormatItem("Last Year of data"),
	new fileFormatItem("Number of Data Years Used"),
	new fileFormatItem("52 Week High Price"),
	new fileFormatItem("52 Week Low Price"),
	new fileFormatItem("Number of Common Shares Outstanding"),
	new fileFormatItem("Number of Authorized Common Shares"),
	new fileFormatItem("Number of Preferred Shares Outstanding"),

	// 100 items

	new fileFormatItem("Number of Authorized Preferred Shares"),
	new fileFormatItem("Total Debt"),
	new fileFormatItem("Potential Dilution"),
	new fileFormatItem("Quarterly sales", QUARTERLY, 2, 0, -4),
	new fileFormatItem("Quarterly E/S", QUARTERLY, 2, 0, -4),
	new fileFormatItem("Last CY of data +5"),
	new fileFormatItem("Outside E/S Estimate"),
	new fileFormatItem("Choice Low Price (5-Year Average Low)"),

	// 110 items

	new fileFormatItem("Choice High PE"),
	new fileFormatItem("Choice Low PE"),
	new fileFormatItem("High Yield"),
	new fileFormatItem("Add/Delete flag", SINGLE, 10, 0, 0),
	new fileFormatItem("Low E/S Estimate"),
	new fileFormatItem("Quarterly E/S Growth"),
	new fileFormatItem("Projected E/S Growth"),
	new fileFormatItem("Quarterly Sales Growth"),
	new fileFormatItem("Projected Sales Growth"),
	new fileFormatItem("Current Yield"),
	new fileFormatItem("Average Yield"),

	// 130 items

	new fileFormatItem("Historic E/S Srowth"),
	new fileFormatItem("Average PE"),
	new fileFormatItem("U:D ratio"),
	new fileFormatItem("5-Year Sales Estimate"),
	new fileFormatItem("Average Pre-Tax Profit"),
	new fileFormatItem("Income Tax Rate"),
	new fileFormatItem("Preferred Dividends"),
	new fileFormatItem("Projected Liabilities"),
	new fileFormatItem("Projected Outstanding Shares"),
	new fileFormatItem("Prefferred Procedure E/S"),

	// 140 items

	new fileFormatItem("Visual E/S Estimate"),
	new fileFormatItem("Last CY Trend E/S"),
	new fileFormatItem("Historic Sales Growth"),
	new fileFormatItem("Average ROE"),
	new fileFormatItem("Low Price Estimate"),
	new fileFormatItem("Average Price Estimate"),
	new fileFormatItem("Sell Level Price"),
	new fileFormatItem("Projected Dividend Payout %"),
	new fileFormatItem("Previous E/S"),
	new fileFormatItem("New E/S Estimate"),

	// 150 items

	new fileFormatItem("Projected High Price"),
	new fileFormatItem("Cash", ANNUAL),
	new fileFormatItem("Accounts Receivables", ANNUAL),
	new fileFormatItem("Inventories", ANNUAL),
	new fileFormatItem("Assets", ANNUAL),
	new fileFormatItem("Short Term Debt (Sum Last 4 Qtrs)", ANNUAL),
	new fileFormatItem("Liabilities", ANNUAL),
	new fileFormatItem("Cash", ANNUAL, 1, -1, 0),
	new fileFormatItem("Accounts Receivables", ANNUAL, 1, -1, 0),
	new fileFormatItem("Inventories", ANNUAL, 1, -1, 0),

	// 160 items

	new fileFormatItem("Assets", ANNUAL, 1, -1, 0),
	new fileFormatItem("Short Term Debt", ANNUAL, 1, -1, 0),
	new fileFormatItem("liabilities", ANNUAL, 1, -1, 0),
	new fileFormatItem("Total Capitalization"),
	new fileFormatItem("Long Term Debt", ANNUAL, 10, -9, 1),
	new fileFormatItem("Number shares outstanding", ANNUAL, 10, -9, 1),
	new fileFormatItem("Cash flow/share", ANNUAL, 10, -9, 1),
	new fileFormatItem("Total Assets"),
	new fileFormatItem("Interest Xoverage %"),
	new fileFormatItem("Total Return"),
	new fileFormatItem("Projected E/S Growth"),
	new fileFormatItem("Severe Low Price"),
	new fileFormatItem("R Squared"),

	// 200 Items

	new fileFormatItem("Reserved"),
	new fileFormatItem("Quarterly Sales", QUARTERLY, 8, -7, 1),
	new fileFormatItem("Quarterly E/S", QUARTERLY, 8, -7, 1),
	new fileFormatItem("Quarterly Dividends", QUARTERLY, 8, -7, 1),
	new fileFormatItem("Inside Ownership"),
	new fileFormatItem("Institutional Holdings"),
	new fileFormatItem("Reserved"),
	new fileFormatItem("ACE 5 Year Estimate E/S Growth Rate"),
	new fileFormatItem("Predictability"),
	new fileFormatItem("Long-Term Debt"),
	new fileFormatItem("Fiscal Year End Month"),
	new fileFormatItem("Reserved", SINGLE, 3, 0, 0),
	new fileFormatItem("Name"),
	new fileFormatItem("Symbol"),
	new fileFormatItem("Exchange"),
	new fileFormatItem("Data Source"),
	new fileFormatItem("Date of Data Creation"),

	// 240 items

	new fileFormatItem("V-L page"),
	new fileFormatItem("Code"),
	new fileFormatItem("Reserved"),
	new fileFormatItem("Industry"),
	new fileFormatItem("Split Ratio During Quarter"),
	new fileFormatItem("Reserved", SINGLE, 9, 0, 0),
	new fileFormatItem("Latest Quarter Ending Date"),
	new fileFormatItem("Reserved", SINGLE, 14, 0, 0),
	new fileFormatItem("Month of Latest Quarterly Data"),

	// 270 items

	new fileFormatItem("Year of Latest Quarterly Data"),
	new fileFormatItem("Reserved", SINGLE, 4, 0, 0),
	new fileFormatItem("Software Version"),
	new fileFormatItem("Reserved", SINGLE, 2, 0, 0),
	new fileFormatItem("SIC Code"),
	new fileFormatItem("S & P Stock Rank"),

	// 280 items

	new fileFormatItem("Financial Strength VL"),
	new fileFormatItem("Safety VL"),
	new fileFormatItem("Reserved", SINGLE, 3, 0, 0),
	new fileFormatItem("Beta"),
	new fileFormatItem("Reserved", SINGLE, 28, 0, 0),
	new fileFormatItem("Quarterly Pre-Tax Income", QUARTERLY, 8, -7, 1),
	new fileFormatItem("Quarterly Net Income", QUARTERLY, 8, -7, 1),

	// 330 items

	new fileFormatItem("Reserved", SINGLE, 16, 0, 0),
	new fileFormatItem("E/S in 5 Yrs. Based on Ace Proj Growth Rate"),
	new fileFormatItem("Historical 5 Yr E/S Growth Rate"),
	new fileFormatItem("Current PE/Ace 5 Yr Growth rate"),
	new fileFormatItem("S&P Rating"),

	// 350 items

	new fileFormatItem("Avg 5 Yr PE"),
	new fileFormatItem("Proj 5 Yr High Price"),
	new fileFormatItem("Proj Compounded Appreciation of Price"),
	new fileFormatItem("Current Yield"),
	new fileFormatItem("Proj Total Return"),
	new fileFormatItem("Historical 5 Yr. Sales Growth Rate"),
	new fileFormatItem("Historical 5 Yr E/S Growth R-Squared"),
	new fileFormatItem("Industry Group"),
	new fileFormatItem("% Off 52 Week High Price"),
	new fileFormatItem("% Above 52 Week Low Price"),

	// 360 items

	new fileFormatItem("Relative Value"),
	new fileFormatItem("PE/Avg ROE"),
	new fileFormatItem("Debt as % of Capital"),
	new fileFormatItem("Next Year E/S estimate"),
	new fileFormatItem("Historical 5 Yr. Sales Growth R-Squared"),
	new fileFormatItem("Reserved", SINGLE, 35, 0, 0)

	// 400 items

    };

    // ****************************************************************
    // Data file I/O routines

    // Data file writing

    // writeSSGfile - write out all data in SSG-1 format to the specified file
    public void writeSSGfile(String fname) {
	create_relative_dates();
	writeData(fname, format_SSG1, "\n", "");
    };

    void create_relative_dates() {
	// ????????????????????????????????????????????????????????????????
	// To be filled in

        // ensure that any relative data attributes are stored for
        // retrieval by the write routines

    }

    // writeData - actually write the data to a file.  Parameters are:
    // fname - the file name to write out
    // items - an array of fileFormatItem that describe what needs to be written out
    // sep - the separator to be written between items
    // ifnull - what should be written out if a value is NULL
    void writeData(String fname, fileFormatItem[] items, String sep, String ifnull) {

	PrintWriter file;
	try {
	    file = new PrintWriter(new FileOutputStream(fname));
	} catch (Exception exc) {
	    System.out.println("Error in opening file " + fname);
	    return;
	}
	for (int i = 0; i < items.length; i++) {
	    writeItem(file, items[i], sep, ifnull);
	}
	file.close();
    }

    // write an individual data attribute to the output file
    void writeItem (PrintWriter file, fileFormatItem item, String sep, String ifnull) {
	String key = item.label;
	String val;

	switch (item.itmType) {
	case CONSTANT:
	    val = key;
	    for (int i = 0; i < item.count; i++) {
		file.print(val + sep);
	    }
	    break;
	case STATIC:
	    val = (String) staticvals.get(key);
	    if (val == null) { val = ifnull; }
	    for (int i = 0; i < item.count; i++) {
		file.print(val + sep);
	    }
	    break;
	case PERIODIC:
	    DataSet ds = (DataSet) datasets.get(key);
	    if (ds == null) {
		val = ifnull;
		for (int i = 0; i < item.count; i++) {
		    file.print(val + sep);
		}
	    } else {
		Enumeration e = ds.getSequence(item.count, item.offset, item.increment);
		while (e.hasMoreElements()) {
		    val = (String) e.nextElement();
		    if (val == null) { val = ifnull; }
		    file.print(val + sep);
		}
	    }
	    break;
	default:
	    System.out.println("Unknown format type " + item.itmType +
			       " for file item " + key);
	}}

    // Data file loading

    // loadSSGfile - read in all data in SSG-1 format from the specified file
    public void loadSSGfile(String fname) {
	loadData(fname, format_SSG1, "\n", "");
	apply_relative_dates();
    }

    void apply_relative_dates() {
        // after the file is loaded, convert all relative data to absolute
        // dates.  relative data loaded includes annual data and quarterly
        // data

	// convert relative annual data to absolute
	try {
	    int lastyr;

	    lastyr = Integer.parseInt(get("Last Year of data"));
	    setBasePeriod(lastyr, 1, 1);
	} catch (Exception exc) {
	    System.out.println("Error: Can't set annual dates.");
	}

	// convert relative quarterly data to absolute
	try {
	    int lastyr, lastqtr, fymonth;

	    lastyr = Integer.parseInt(get("Year of Latest Quarterly Data"));
	    lastqtr = Integer.parseInt(get("Month of Latest Quarterly Data"));
	    fymonth = Integer.parseInt(get("Fiscal Year End Month"));

	    // convert reported and FY months to quarters (range 0 to 3)
	    int fyendqtr = fymonth / 3;
	    int mostrecentqtr = lastqtr / 3;

	    // convert reported quarter to FY quarter (range 1 - 4) based on FY
	    // end quarter
	    int fyqtrreported = 1 + (4 + mostrecentqtr - fyendqtr - 1) % 4;

	    setBasePeriod(lastyr, fyqtrreported, 4);
	} catch (Exception exc) {
	    System.out.println("Error: Can't set quarterly dates");
	}
    };

    // load input data from a file
    void loadData(String fname, fileFormatItem[] items,
		  String sep, String ifnull) {

	BufferedReader file;
	try {
	    FileInputStream fis = new FileInputStream(fname);
	    file = new BufferedReader(new InputStreamReader(fis));
	} catch (Exception exc) {
	    System.out.println("Error in opening file " + fname);
	    return;
	}
	for (int i = 0; i < items.length; i++) {
	    try {
		loadItem(file, items[i], sep, ifnull);
	    } catch (Exception exc) {
		System.out.println("Error: " + exc);
		return;
	    }
	}
	try {
	    file.close();
	} catch (Exception exc) {
	}
    }

    void loadItem (BufferedReader file, fileFormatItem item,
		   String sep, String ifnull)
	throws IOException {
	String key = item.label;
	String val;
	switch (item.itmType) {
	case CONSTANT:
	    for (int i = 0; i < item.count; i++) {
		val = file.readLine();
	    }
	    break;
	case STATIC:
	    for (int i = 0; i < item.count; i++) {
		val = file.readLine();
		if (! val.equals("")) {
		    set(val, key);
		}
	    }
	    break;
	case PERIODIC:
	    int period = item.offset;
	    int incr = item.increment;
	    for (int i = 0; i < item.count; i++) {
		val = file.readLine();
		if (! val.equals("")) {
		    set(val, key, 0, period, item.frequency);
		}
		period += incr;
	    }
	    break;
	default:
	    System.out.println("Unknown format type " + item.itmType +
			       " for file item " + key);
	}}
}
