
// A DataSet is a collection of data for a set of periods.
// e.g.
//     Sales in 1998 Q1 = 54,
//     Sales in 1997 Q2 = 83, etc.

// Each DataSet knows which is the latest & earliest periods stored.


package sptossg ;

import java.util.*;

public class DataSet {

    // ****************************************************************
    // Data structures initialization & processing

    Period maxper = null;
    Period minper = null;
    int frequency = 0;
    Hashtable vals = new Hashtable();
    String name;

    Period baseperiod = null;


    Period baseper = null;

    public DataSet (String name) {
	this.name = name;
    }

    public int frequency() {
	return frequency;
    }

    // getSequence - return a sequence of length COUNT of the values,
    // starting at the maximum period adjusted by OFFSET, and stepping
    // through the periods by INCREMENT.
    //
    // e.g. getSequence(8, -12, 1)
    //   if this is a quarterly data item, this will start three years ago
    //   (-12 quarters), and return an 8 element sequence covering every
    //   quarter (incrementing by 1) in the two years following
    public Enumeration getSequence(int count, int offset, int increment) {
	Vector itms = new Vector();
	Period per = maxper.adjust(offset);
	for (int i = 0; i < count; i++) {
	    itms.addElement(retrieve(per));
	    per = per.adjust(increment);
	}
	return itms.elements();
    }

    public void setBasePeriod(Period baseperiod) {
	this.baseperiod = baseperiod;
	// loop through all existing values, normalizing relative items;
	Enumeration e = vals.keys();
	while (e.hasMoreElements()) {
	    Period key = (Period) e.nextElement();
	    String value = (String) vals.get(key);
	    if (key.year() == 0) {
		Period newp = baseperiod.adjust(key.period());
		vals.remove(key);
		store(value, newp);
	    }
	}
    }

    // ****************************************************************
    // Storage methods

    public void store(String s, int year, int yearpart, int frequency) {
	Period p = new Period(year, yearpart, frequency);
	store(s, p);
    }

    public void store(String s, String year, String yearpart) {
	Period p = new Period(year, yearpart);
	store(s, p);
    }

    public void store(String s, Period p) {
	if (frequency == 0) {
	    frequency = p.frequency();
	} else if (frequency !=  p.frequency) {
	    // throw new incompatibleFormatException();
	    System.out.println("Error: incompatible period type. ");
	    return;
	}
	vals.put(p, s);
	// save max & min period (only if doing absolute dating)
	if (p.year() != 0) {
	    if (p.greater(maxper)) maxper = p;
	    if (p.less(minper)) minper = p;
	}
	return;
    }

    // ****************************************************************
    // Retrieval methods

    // retrieval by strings; e.g. retrieve("1997", "Q3")
    public String retrieve(String year, String yearpart) {
	Period p = new Period(year, yearpart);
	return retrieve(p);
    }

    // retrieval by period number; e.g. retrieve(1997, 3, 4)
    public String retrieve(int year, int yearpart, int frequency) {
	Period p = new Period(year, yearpart, frequency);
	return retrieve(p);
    }

    // retrieval by period class instance; e.g. retrieve(new period("1997"))
    public String retrieve(Period p) {
	String result = (String) vals.get(p);
	// when year == 0, then it's a relative offset (e.g. -3 periods)
	if (result == null && p.year() == 0) {
	    return retrieve(p.period());
	}
	return result;
    }

    // retrieval by offset; e.g. retrieve(-3) gets the value for the third
    // period back from the base period.
    public String retrieve(int offset) {
	if (baseperiod == null) return null;
	return retrieve(baseperiod.adjust(offset));
    }




    // ****************************************************************
    // Routines for testing

    // display all values stored in the data set.
    public void dump(CompanyStorage company, String keystr) {
	boolean dorounding = false;
	if (minper == null && maxper == null ) {
	    dorounding = true;
	    System.out.println("    " + keystr + ":  (Dynamic DataSet)");
	} else {
	    System.out.println("    " + keystr + ":  <" +
			       minper + " - " + maxper + ">");
	}
	Vector keys = getKeys(company);
	Enumeration e = keys.elements();
	while (e.hasMoreElements()) {
	    Object key = e.nextElement();
	    String val = retrieve((Period) key);
	    if (dorounding) {
		double nval = (new Double(val)).doubleValue();
		int nval2 = (int) (nval * 100);
		val = "" + (nval2 / 100.0);
	    }
	    if (val != null && ! val.equals("")) {
		System.out.println("        " + key + ": " + val);
	    } else {
		System.out.println("Got null for " + key);
	    }
	}
    }

    // getKeys - Return all keys for a DataSet, in sorted order
    public Vector getKeys(CompanyStorage Company) {
	Hashtable copy = (Hashtable) vals.clone();
	Vector keys = new Vector();

	while (! copy.isEmpty()) {
	    Enumeration e = copy.keys();
	    Period min = (Period) e.nextElement();
	    while (e.hasMoreElements()) {
		Period per = (Period) e.nextElement();
		if (per.less(min)) { min = per; }
	    }
	    keys.addElement(min);
	    copy.remove(min);
	}
	return keys;
    }

}
