
// Class DynamicDataSet

// This class defines datasets whose values can be calculated.
//  e.g.
//     Outstanding Shares <=> "Net Profits" divided by "EPS"

// This class extends the DataSet class as follows:

// The constructor takes a Formula for use in calculating values

// The semantics of the storage methods are the same as in DataSet and store
// values explicitly.

// The semantics of retrieval is that if an explicit value is stored, then
// that will be used.  If no explicit value is available, then the formula is
// used to calculate the result value.

// Usage:
//    The constructors of DynamicDataSet are private methods and not
//    accessible to other classes.  Instead, there is a public class method
//    'define(name, formula-definition)' to be used.
//    Example Usage:
//        DynamicDataSet.define("PTP Margin", "100 * [PreTax Inc] / Sales");

// Also, DynamicDataSets are not defined on a per-company basis, but
// instead are defined generally.  Retrievals of a DDS for a company will
// instantiate that DDS within that company the first time a retrieval is
// attempted.

package sptossg ;

import java.util.*;

public class DynamicDataSet extends DataSet {

    CompanyStorage company;
    Formula formula;
    String name;

    static Hashtable allGenericSets = new Hashtable();

    // Constructor for a generic DynamicDataSet (one not yet attached to a
    // company)
    private DynamicDataSet (String name, Formula formula) {
	super(name);
	this.name = name;
	this.formula = formula;
	allGenericSets.put(name, this);
    }

    // Constructor for a specific DynamicDataSet (one attached to a company)
    private DynamicDataSet (String name, Formula formula, CompanyStorage company) {
	super(name);
	this.name = name;
	this.formula = formula;
	this.company = company;
    }

    static public DynamicDataSet define(String name, String def) {

	Formula formula = Formula.parse(def);
	if (formula == null) { return null; }
	return new DynamicDataSet(name, formula);
    }

    // instantiate - copy an uninstantiated DDS for use with a specific company
    public DynamicDataSet instantiate(CompanyStorage company) {
	DynamicDataSet newDS = new DynamicDataSet(name, formula, company);
	return newDS;
    }

    // static method to find a named, uninstantiated DynamicDataSet
    static public DynamicDataSet find(String name) {
	Object set = allGenericSets.get(name);
	if (set == null) { return null; }
	return (DynamicDataSet) set;
    }

    // doing_evaluation is used for automatic cycle detection during
    // evaluation.  For example, if A = B / 10 and B = A * 10, an
    // evaluation of either A or B will cause a undesired cycle.
    boolean doing_evaluation = false;

    // supercede DataSet's retrieve(Period p)
    //  if not explicitly stored in DataSet, then use formula to get value
    public String retrieve(Period p) {
	String result = super.retrieve(p);
	double val;

	if (result == null) {
	    // if we're already in an evaluation of this dynamic attribute,
	    // then terminate the cycle with no value.  Note that this
	    // cycle detection should allow us to define DDS formulas based
	    // on one another and do retrievals as long as one of the
	    // attributes has an explicit value stored.
	    if (doing_evaluation) {
		System.out.println("Warning: Evaluation cycle detected on " + name);
		val = Double.NaN;
	    } else {
		try {
		    doing_evaluation = true;
		    val = formula.evaluate(company, p);
		} finally {
		    doing_evaluation = false;
		}
	    }
	    if (Double.isNaN(val)) { result = null; }
	    else { result = ("" + val); }
	}
	return result;
    }

    // getKeys - Return all keys , in sorted order
    public Vector getKeys(CompanyStorage company) {
	return Formula.merge(super.getKeys(company), formula.getKeys(company));
    }

    // this is a retrieval method for a static values formula
    public String calculate() {
	// ????????????????????????????????????????????????????????????????
	// TBD
	return "0";
    }
}

