
// Class Formula

// The Formula class implements a small calculations language, used to
// define formulas to be calculated.  Formulas are associated with data
// attributes (via the DynamicDataSet class), and perform mathematical
// operations based on other data attributes and constants.

// Examples:
//    Pre-Tax Profit Margin:
//                100 * "PreTax Income" / "Sales"
//    Last FY Quarter Data:
//                1 + (4 + ("Month of Latest Quarterly Data" / 3)
//                       - ("Fiscal Year End Month" / 3) - 1) % 4

// Formulas are written as standard mathematical equations,
//   e.g. x + 3, 2 * y, 4 * (a1 - a2)

// To reference another attribute, you can just put the attribute name into
// the formula, e.g. Sales / 1000

// If the attribute name contains spaces or other non-alphanumeric
// characters, then you must surround the attribute name with square
// brackets, e.g. [Earnings per Share], [Sales (in millions)], etc., though
// you can always use brackets even when not required.

// Functions can be put in formulas in standard C/C++/Java syntax, 
// e.g. 2 * max([Long Term Debt], [Short Term Debt])

// Available operators/functions for Formulas are:
//  Binary operators
//    +		- addition
//    -		- subtraction
//    *		- multiplication
//    /		- division
//    %		- remainder
//    min	- minimum
//    max	- maximum

// Unary operators
//    abs	- absolute value

// Other
//    first - first available, e.g. first("Total Debt", "Long-Term Debt")
//              will return the total debt for the specified period, if
//              available.  If not available, then Long-Term Debt is
//              checked next.
//
//    prev - previous period's value for attribute;
//		examples: 
//	            Sales - prev(Sales) 	
//				the increase in Sales from year to year
//		    100 * Sales / prev(Sales)
//				the % increase in Sales from year to year
//      prev can also take an optional second parameter to calculate
//	the no of periods difference, e.g.
//		100 * [Quarterly EPS] / prev([Quarterly EPS], 4)
//				calculates the year ago percentage change
//				in the quarterly earnings per share

package sptossg ;

import java.util.*;
import java.io.*;

public class Formula {

    String operatorStr;
    int operator;
    Stack params;
    String attribute;
    double value;


    // Constructors:


    // Binary operator constructor, e.g. A + B, mod(x, 4)
    public Formula (Formula left, String operator,  Formula right) {
	params = new Stack();
	params.push(right);	
	params.push(left);
	initSelf(operator, params);
    }

    // Unary operator constructor, e.g. abs(a)
    public Formula (String operator, Formula left) {
	params = new Stack();
	params.push(left);
	initSelf(operator, params);
    }

    // N-ary operator constructor, e.g. max(a,b,c,d)
    public Formula (String operator,  Stack params) {
	initSelf(operator, params);
    }

    int ParseOp(String operator) {
	Integer val = (Integer) OpsCodes.get(operator);

	if (val == null) return 0;
	return val.intValue();
    }

    // actual initialization method for function operators
    private void initSelf(String operator, Stack params) {
	Integer val = (Integer) OpsCodes.get(operator);
	if (val == null) return;
	
	this.operator = val.intValue();
	this.operatorStr = operator;
	this.params = params;
    }

    // constructor for reference to a data attribute
    public Formula(String attr) {
	this.attribute = attr;
	this.operator = FETCH;
	this.operatorStr = "FETCH";
	params = null;
    }

    // constructor for reference to a constant
    public Formula(double value) {
	this.value = value;
	this.operator = CONSTANT;
	this.operatorStr = "CONSTANT";
	params = null;
    }

    public String toString() {
	if (operatorStr.equals("CONSTANT")) {
	    // return("" + value);
	    return("const(" + value + ")");
	} else if (operatorStr.equals("FETCH")) {
	    return("get(\"" + attribute + "\")");
	} else if (params.size() == 1) {
	    Formula param = (Formula) params.peek();
	    return(operatorStr + "(" + param + ")");
	} else if (operatorStr.length() == 1 && params.size() == 2) {
	    Stack nparams = (Stack) params.clone();
	    Formula f1 = (Formula) nparams.pop();
	    Formula f2 = (Formula) nparams.pop();
	    return("(" + f1 + " " + operatorStr + " " + f2 + ")");
	} else {
	    Stack nparams = (Stack) params.clone();
	    String sep = "(";
	    String paramstr = "";
	    while (! nparams.empty()) {
		paramstr += sep + nparams.pop();
		sep = ", ";
	    }
	    return (operatorStr + paramstr + ")");
	}}

    static Hashtable OpsCodes = new Hashtable();
    static Hashtable OpsMin = new Hashtable();
    static Hashtable OpsMax = new Hashtable();
    static Hashtable OpsEval = new Hashtable();

    static void DefineOp(String opstr, int code, int minargs,
			 int maxargs, int evalCount) {
	Integer icode = new Integer(code);

	if (opstr != null) {
	    OpsCodes.put(opstr, icode);
	}
	OpsMin.put(icode, new Integer(minargs));
	OpsMax.put(icode, new Integer(maxargs));
	OpsEval.put(icode, new Integer(evalCount));
    }

    // Define integer constants for available functions and internal operators

    // internal operators
    static public final int FETCH = 1;
    static public final int CONSTANT = 2;

    // special operators - evaluation of params is handled differently
    static public final int PREV = 11;
    static public final int FIRST = 12;

    // regular operators (unary, binary, N-ary)
    static public final int ABS = 101;
    static public final int PLUS = 103;
    static public final int MINUS = 104;
    static public final int TIMES = 105;
    static public final int DIVIDE = 106;
    static public final int REMAINDER = 107;
    static public final int MIN = 108;
    static public final int MAX = 109;

    // Define all built-in functions and internal operators
    static {
	// DefineOp - takes five params:
	//    function name - string name for parsing
	//    function code - integer representation
	//    min and max no of params
	//    how many params to evaluate - usually the max, except
	//	for special functions

	DefineOp(null, FETCH, 0, 0, 0);
	DefineOp(null, CONSTANT, 0, 0, 0);

	DefineOp("prev", PREV, 1, 1, 0);
	DefineOp("first", FIRST, 2, 10, 0);

	DefineOp("abs", ABS, 1, 1, 1);

	DefineOp("+", PLUS, 1, 2, 2);
	DefineOp("-", MINUS, 1, 2, 2);
	DefineOp("*", TIMES, 2, 2, 2);
	DefineOp("/", DIVIDE, 2, 2, 2);
	DefineOp("%", REMAINDER, 2, 2, 2);
	DefineOp("min", MIN, 2, 10, 10);
	DefineOp("max", MAX, 2, 10, 10);

    }


    // ****************************************************************
    // code to parse string representations of a formula
    //  e.g. "Sales / 1000", "100 * [PreTax Inc] / Sales", 
    //    "max(Sales - Expenses, 0)"

    // Formula scanner - breaks up a formula string into a Stack of tokens
    static Stack tokenize(String formulastr) {
	StringTokenizer st = new StringTokenizer(formulastr, " ", true);
	Stack tokens = new Stack();
	while (st.hasMoreTokens()) {
	    String token = nextToken(st);
	    tokens.addElement(token);
	}

        // reverse tokens so that first tokens are at top of Stack and add
        // an end-marker.
	Stack result = new Stack();
	result.push("EOF");
	while (! tokens.empty()) {
	    result.push(tokens.pop());
	}
	return result;
    }

    // Scan the string to return the next token.  A token is:
    //   An operator - A special character - e.g. "+", "-", "*", "/", "%", etc.
    //   A number - a sequence of digits
    //   An Attribute - a sequence of characters, starting with an alphabetic.

    //      Note: if an attribute has any spaces or special characters
    //      contained within, then it can be surrounded by square brackets
    //      to denote the entire unit,
    //      e.g. [Net Profit], [Sales (in millions)].  The scanner will
    //      strip off the brackets before returning the token.
    static String nextToken(StringTokenizer st) {
	String sepchars = "!@#$%^&*()-=+[{]};:',/<>? \t";
	String token = null;
	while (token == null) {
	    token = st.nextToken(sepchars);
	    if (token.trim().length() == 0) {
		token = null;
	    } else {
		if (token.equals("[")) {
		    token = st.nextToken("]");
		    st.nextToken(sepchars); // skip the ']'
		}
	    }
	}
	return token;
    }

    // ****************************************************************
    // parse(<FORMULA STRING>) - A recursive descent parser for formulas.
    // Formulas can be described by the following grammar:

    // E           -> E + E | E - E | E + E | E / E | E % E | ( E ) |
    //		      ATOM | NUMBER | FN-CALL
    // FN-CALL     -> ATOM ( PARAM-LIST )
    // PARAM-LIST  -> E | E , PARAM-LIST

    // However, recursive descent parsers require an LL(1) grammar.  The following
    // is the LL(1) grammar used for the formula language:

    // 	S	E1 EOF

    // 	E1	E2 E1Rest

    // 	E1Rest	"+" E2 E1Rest
    // 		"-" E2 E1Rest
    // 		<null>

    // 	E2	ATOM E2Rest

    // 	E2Rest	"*" ATOM E2Rest
    // 		"/" ATOM E2Rest
    // 		"%" ATOM E2Rest
    // 		<null>

    // 	ATOM	NUMBER
    // 		"(" E1 ")"
    // 		OPERAND "(" PARAMS
    // 		OPERAND

    // 	PARAMS	E1 ")"
    // 		E1 "," PARAMS ")"

    // The parser for this grammar is implemented as a recursive descent
    // parser.  All non-terminals in the grammar are parsed via their own
    // parsing routines, named parse_<NON-TERMINAL>.  Each parse routine
    // takes the stack of tokens and returns a Formula representing that
    // sub-expression.  In the cases where a non-terminal does not
    // represent a complete sub-expression (E1Rest & E2Rest), the parse
    // routine takes an additional parameter which completes the formula.


    public static Formula parse(String formulastr) {
	Stack toks = tokenize(formulastr);
	Formula f;

	execution = new Stack();
	try {
	    f = parse_S(toks);
	} catch (Exception exc) {
	    System.out.println("Error on parse: " + exc);
	    dumpstack(execution,"Execution Stack");
	    return null;
	}
	if (f == null) { return null; }

	return f;
    }

    static Stack execution;
    static void enter(String name) {
	execution.push(name);
	// System.out.println("Entering " + name);
    }

    static void exit(String name) {
	String old = (String) execution.pop();
// 	if (name.equals(old)) {
// 	    System.out.println("Exiting " + name);
// 	} else {
// 	    System.out.println("Exiting " + name + ", but stack think its " + old);
// 	}
    }

    static void parse_error(Stack tokens, String node) {
	System.out.println("Parse Error: " + node);
    }

    // Parser for S
    // 	S	E1 EOF
    static Formula parse_S(Stack tokens) {
	enter("parse_S");
	Formula f = parse_E1(tokens);
	if (f == null) { return null; }
	String token = (String) tokens.pop();
	if (token.equals("EOF") && tokens.empty()) {
	    exit("parse_S");
	    return f;
	}
	parse_error(tokens, "parse_S");
	return null;
    }

    // Parser for E1
    // 	E1	E2 E1Rest
    static Formula parse_E1(Stack tokens) {
	enter("parse_E1");
	Formula f1 = parse_E2(tokens);
	if (f1 == null) { return null; }
	Formula f = parse_E1Rest(tokens, f1);
	if (f != null) {
	    exit("parse_E1");
	    return f;
	}
	parse_error(tokens, "parse_E1");
	return null;
    }

    // Parser for E1Rest
    // 	E1Rest	"+" E2 E1Rest
    // 		"-" E2 E1Rest
    // 		<null>
    static Formula parse_E1Rest(Stack tokens, Formula f1) {
	enter("parse_E1Rest");
	String op = (String) tokens.pop();
	Formula f2 = null;
	Formula f3 = null;

	if (op.equals("+") || op.equals("-")) {
	    f2 = parse_E2(tokens);
	}

	if (f2 == null) {
	    tokens.push(op);
	    exit("parse_E1Rest");
	    return f1;
	}

	f3 = parse_E1Rest(tokens, new Formula(f1, op, f2));
	exit("parse_E1Rest");
	return f3;
    }

    // Parser for E2
    // 	E2	ATOM E2Rest
    static Formula parse_E2(Stack tokens) {
	enter("parse_E2");
	Formula f1 = parse_ATOM(tokens);
	if (f1 == null) { return null; }
	Formula f2 = parse_E2Rest(tokens, f1);
	exit("parse_E2");
	return f2;
    }

    // Parser for E2Rest
    // 	E2Rest	"*" ATOM E2Rest
    // 		"/" ATOM E2Rest
    // 		"%" ATOM E2Rest
    // 		<null>
    static Formula parse_E2Rest(Stack tokens, Formula f1) {
	enter("parse_E2Rest");
	String op = (String) tokens.pop();
	Formula f2 = null;
	Formula f3 = null;

	if (op.equals("*") || op.equals("/") || op.equals("%")) {
	    f2 = parse_ATOM(tokens);
	}

	if (f2 == null) {
	    tokens.push(op);
	    exit("parse_E2Rest");
	    return f1;
	}

	f3 = parse_E2Rest(tokens, new Formula(f1, op, f2));
	exit("parse_E2Rest");
	return f3;
    }

    // 	ATOM	NUMBER
    // 		"(" E1 ")"
    // 		OPERAND "(" PARAMS
    // 		OPERAND
    static Formula parse_ATOM(Stack tokens) {
	enter("parse_ATOM");
	Formula f1;
	String token = (String) tokens.pop();

	if (tokenIsNumber(token)) {
	    f1 = new Formula((new Double(token)).doubleValue());
	    exit("parse_ATOM");
	    return f1;
	}
	if (token.equals("(")) {
	    f1 = parse_E1(tokens);
	    token = (String) tokens.pop();
	    if (token.equals(")")) {
		exit("parse_ATOM");
		return f1;
	    }
	    parse_error(tokens, "parse_ATOM-1");
	    return null;
	}
	if (tokenIsOperand(token)) {
	    String next = (String) tokens.pop();
	    if (next.equals("(")) {
		Stack params = parse_PARAMS(tokens);

		// verify function in function call
		Integer code = (Integer) OpsCodes.get(token);

		if (code == null) {
		    parse_error(tokens, "Function error: " +
				"unknown function " + token);
		    return null;
		}

		// verify parameter count for function
		int minparams;
		int maxparams;
		try {
		    minparams = ((Integer) OpsMin.get(code)).intValue();
		    maxparams = ((Integer) OpsMax.get(code)).intValue();
		} catch (Exception exc) {
		    System.out.println("Error: " + exc);
		    return null;
		}

		int size = params.size();

		if (size < minparams || size > maxparams) {
		    parse_error(tokens, "Function error: " +
				size + " is invalid number of params" +
				" [" + minparams + "," + maxparams + "] " +
				", function "
				+ token);
		    return null;
		}

		f1 = new Formula(token, params);
		exit("parse_ATOM");
		return f1;
	    }
	    tokens.push(next);

	    f1 = new Formula(token);
	    exit("parse_ATOM");
	    return f1;
	}
	parse_error(tokens, "parse_ATOM-2");
	return null;
    }

    // Parser for PARAMS
    // 	PARAMS	E1 ")"
    // 		E1 "," PARAMS ")"
    //

    // Note: this is the one inconsistent parse routine in that it doesn't
    // return a Formula like all the others, but instead returns a Stack of
    // Formulas.
    static Stack parse_PARAMS(Stack tokens) {
	enter("parse_PARAMS");
	Stack formulas;

	Formula f1 = parse_E1(tokens);
	if (f1 == null) {
	    parse_error(tokens, "parse_PARAMS-1");
	    return null;
	}
	String sep = (String) tokens.pop();
	if (sep.equals(")")) {
	    formulas = new Stack();
	} else if (sep.equals(",")) {
	    formulas = parse_PARAMS(tokens);
	} else {
	    parse_error(tokens, "parse_PARAMS-2");
	    return null;
	}
	formulas.push(f1);
	exit("parse_PARAMS");
	return formulas;
    }

    // a token is numeric if the first character is numeric, though all
    // characters should be numeric
    static boolean tokenIsNumber(String token) {
	char c1 = token.charAt(0);
	if (c1 >= '0' && c1 <= '9') { return true; }
	if (c1 == '.') { return true; }
	return false;
    }

    // a token is an operand if the first character is alphabetic.
    // Subsequent characters can be of any type.
    static boolean tokenIsOperand(String token) {
	char c1 = token.charAt(0);
	if (c1 >= 'a' && c1 <= 'z') { return true; }
	if (c1 >= 'A' && c1 <= 'Z') { return true; }
	return false;
    }

    // an operator is generally a single character (not enforced), and not
    // a number or operand.
    static boolean tokenIsOperator (String token) {
	return ((! tokenIsOperand(token)) && (! tokenIsNumber(token)));
    }

    // debugging routine to dump a stack
    static void dumpstack(Stack toks, String title) {
	System.out.println(title);

	// clone the stack so that we don't modify it.
	Stack s1 = (Stack) toks.clone();
	int i = 0;

	while (! s1.empty()) {
	    String val = (String) s1.pop();
	    System.out.println(i++ + ": " + val);
	}
    }

    // Merge two sorted vectors of Periods into a new sorted vector
    static public Vector merge(Vector v1, Vector v2) {
	if (v1 == null) { return v2; }
	if (v2 == null) { return v1; }

	Vector result = new Vector();

	while ((! v1.isEmpty()) && (! v2.isEmpty())) {
	    Period p1 = (Period) v1.firstElement();
	    Period p2 = (Period) v2.firstElement();

	    // merge vectors, eliminating duplicates
	    if (p1.equals(p2)) {
		result.addElement(p1);
		v1.removeElementAt(0);
		v2.removeElementAt(0);
	    } else if (p1.less(p2)) {
		result.addElement(p1);
		v1.removeElementAt(0);
	    } else {
		result.addElement(p2);
		v2.removeElementAt(0);
	    }
	}

        // One vector was cleaned out; copy remaining elements from the
        // other vector
	Vector v4 = v1;
	if (v1.isEmpty()) { v4 = v2; } 

	while (! v4.isEmpty()) {
	    result.addElement(v4.firstElement());
	    v4.removeElementAt(0);
	}

	return result;
    }

    // get all Period keys for a Formula; This is the *potential* domain
    // for the Formula, calculated as the union of the domain of periods
    // for each attribute referenced within.  Because this is a union, and
    // due to the vagaries of the formula execution, some of these keys may
    // actually not be part of the domain of this Formula.
    public Vector getKeys(CompanyStorage company) {
	if (operator == FETCH) { return company.getKeys(attribute); }

	if (params == null) { return null; }

	Vector keys = null;

	Stack nparams = (Stack) params.clone();
	while (! nparams.empty()) {
	    Formula f1 = (Formula) nparams.pop();
	    keys = merge(keys, f1.getKeys(company));
	}

	return keys;
    }

    // ****************************************************************
    // Formula evaluation routines

    // evaluation for non-date formulas; this probably doesn't work yet due
    // to issues elsewhere.
    public double evaluate(CompanyStorage company) {
	return evaluate(company, null);
    }

    // evaluate a formula within the context of a company, for a particular
    // date.  Method returns a double as a value, or the constant
    // Double.NaN if the formula doesn't have a value for that date.
    public double evaluate(CompanyStorage company, Period date) {

	if (company == null) {
	    System.out.println("Error: no company passed for formula " +
			       this.toString());
	    return Double.NaN;
	}

	double args[] = new double[0];
	int argctr = 0;

	Integer evct = (Integer) OpsEval.get(new Integer(operator));
	Integer maxct = (Integer) OpsMax.get(new Integer(operator));

	Debug.debug("Evaluating " + this.toString());
	
	if (evct == null || maxct == null) {
	    return Double.NaN;
	}
	int evalcount = evct.intValue();
	int maxcount = maxct.intValue();

	// evaluate parameter stack
	if (params != null) {
	    args = new double[params.size()];
	    Stack nparams = (Stack) params.clone();
	    while (! nparams.empty() && evalcount > 0) {
		evalcount--;
		Formula f1 = (Formula) nparams.pop();
		double val = f1.evaluate(company, date);
		if (Double.isNaN(val)) { return Double.NaN; }
		args[argctr++] = val;
	    }
	}

	// build a copy of parameters for fns that don't evaluate all args
	Stack nparams = null;
	if (params != null && evalcount < maxcount) {
	    nparams = (Stack) params.clone();
	}

	double result;
	switch (operator) {
	case FETCH:
	    String valstr = null;

	    // if date given, do a date-specific retrieval
	    if (date != null) {
		valstr = company.get(attribute, date);
	    }

            // if no date given or no value found for date, then see if
            // it's a non-date-specific attribute.
	    if (valstr == null) {
		valstr = company.get(attribute);
	    } 

	    // if no value found, signal error
	    if (valstr == null) {
		return Double.NaN;
	    }

	    // if value found isn't numeric, signal error
	    try {
		result = (new Double(valstr)).doubleValue();
	    } catch (Exception exc) {
		return Double.NaN;
	    }
	    break;
	case CONSTANT:
	    result = value;
	    break;
	case FIRST:
	    result = Double.NaN;
	    while (! nparams.empty()) {
		Formula f1 = (Formula) nparams.pop();
		double val = f1.evaluate(company, date);
		if (! Double.isNaN(val)) {
		    result = val;
		    break;
		}
	    }
	    break;
	case PREV:
	    Formula f1 = (Formula) nparams.pop();

	    int offset;

	    if (! nparams.empty()) {
		Formula f2 = (Formula) nparams.pop();
		offset = (int) f2.evaluate(company, date);
	    } else {
		offset = 1;
	    }

	    Period newdate = date.adjust(- offset);

	    result = f1.evaluate(company, newdate);
	    break;
	    
	case ABS:
	    result = Math.abs(args[0]);
	    break;
	case PLUS:
	    if (argctr == 1) {
		result = args[0];
	    } else {
		result = args[0] + args[1];
	    }
	    break;
	case MINUS:
	    if (argctr == 1) {
		result = 0 - args[0];
	    } else {
		result = args[0] - args[1];
	    }
	    break;
	case TIMES:
	    result = args[0] * args[1];
	    break;
	case DIVIDE:
	    result = args[0] / args[1];
	    break;
	case REMAINDER:
	    result = args[0] % args[1];
	    break;
	case MIN:
	    result = Math.min(args[0], args[1]);
	    break;
	case MAX:
	    result = Math.max(args[0], args[1]);
	    break;
	default:
	    System.out.println("Unknown operator " + operator);
	    return Double.NaN;
	}

	Debug.debug("Evaluation of " + this.toString() + " = " + result);
	return result;
    }
}
